package net.knarcraft.serverlauncher.profile;

import net.knarcraft.serverlauncher.server.Server;
import net.knarcraft.serverlauncher.server.ServerType;
import net.knarcraft.serverlauncher.userinterface.ServerTab;
import net.knarcraft.serverlauncher.Main;

import javax.swing.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.concurrent.Executors;

/**
 * Contains all user settings, and a list of servers.
 *
 * @author		Kristian Knarvik <kristian.knarvik@knett.no>
 * @version	0.0.0.1
 * @since		0.0.0.1
 */
public class Profile {
    private static final ArrayList<Profile> profiles = new ArrayList<>();
    private static Profile current;

    private final ArrayList<Collection> collections;
    private final String name;
    private boolean runInBackground;
    private int delayStartup;
    private boolean downloadJars;

    private Profile(String name) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = false;
        this.delayStartup = 0;
        this.downloadJars = false;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    private Profile(String name, boolean runInBackground, int delayStartup, boolean downloadJars) {
        this.collections = new ArrayList<>();
        this.name = name;
        this.runInBackground = runInBackground;
        this.delayStartup = delayStartup;
        this.downloadJars = downloadJars;
        profiles.add(this);
        if (current == null) {
            current = this;
        }
    }

    public boolean getRunInBackground() {
        return this.runInBackground;
    }

    public int getDelayStartup() {
        return this.delayStartup;
    }

    public boolean getDownloadJars() {
        return this.downloadJars;
    }

    public static Profile getCurrent() {
        return current;
    }

    public ArrayList<Collection> getCollections() {
        return this.collections;
    }

    /**
     * Gets a Collection object by name.
     *
     * @param name  The name of the collection.
     * @return      A collection object.
     */
    public Collection getCollection(String name) {
        for (Collection collection : this.collections) {
            if (collection.getName().equals(name)) {
                return collection;
            }
        }
        return null;
    }

    public static Profile getProfile(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                return profile;
            }
        }
        return null;
    }

    public void setRunInBackground(boolean value) {
        this.runInBackground = value;
    }

    public void setDelayStartup(int value) {
        if (value >= 0) {
            this.delayStartup = value;
        }
    }

    public void setDownloadJars(boolean value) {
        this.downloadJars = value;
    }

    public static  void setCurrent(String name) {
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                current = profile;
                break;
            }
        }
    }

    /**
     * Adds a collection to the profile if the name is valid.
     *
     * @param name  The name of the collection and its elements.
     */
    public void addCollection(String name) {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (getCollection(name) == null &&
                !name.equals("") &&
                !name.equals("All") &&
                !name.contains("!") &&
                !name.contains("?") &&
                !name.contains(";")
                ) {
            collections.add(new Collection(name));
        } else {
            JOptionPane.showMessageDialog(
                    null,
                    "A server name must my unique and not empty or \"All\"." +
                            "It can't contain any of the characters \"!\", \"?\" or \";\".",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
        }
    }

    /**
     * Adds a profile if the name is valid and unique.
     *
     * @param name  The name of the new profile.
     */
    public static void addProfile(String name) {
        if (name == null) { //If a user cancels or crosses out window
            return;
        }
        if (name.equals("") && !name.contains("!") && !name.contains("?") && !name.contains(";")) {
            JOptionPane.showMessageDialog(
                    null,
                    "Profile name can't be blank.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
            return;
        }
        for (Profile profile : profiles) {
            if (profile.name.equals(name)) {
                JOptionPane.showMessageDialog(
                        null,
                        "There is already a profile with this name.",
                        "Error",
                        JOptionPane.ERROR_MESSAGE
                );
                return;
            }
        }
        new Profile(name);
        Main.gui().addProfile(name);
    }

    public void removeCollection(String name) {
        for (int i = 0; i < collections.size(); i++) {
            if (collections.get(i).getName().equals(name)) {
                this.collections.remove(i);
                Main.gui().update();
                break;
            }
        }
    }

    public static void removeProfile(String name) {
        if (profiles.size() > 1) {
            for (int i = 0; i < profiles.size(); i++) {
                if (profiles.get(i).name.equals((name))) {
                    profiles.remove(i);
                    Main.gui().removeProfile(i);
                }
            }
        }
    }

    /**
     * Sends a command to a server, or all servers
     *
     * @param serverName    The target server
     * @param command       The command to send.
     */
    public void sendCommand(String serverName, String command) {
        if (serverName.equals("All")) {
            for (Collection collection : this.collections) {
                try {
                    collection.getServer().sendCommand(command);
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(
                            null,
                            "Server " + collection.getName() + " caused an exception.",
                            "Error",
                            JOptionPane.ERROR_MESSAGE
                    );
                }
            }
        } else {
            Collection collection = getCollection(serverName);
            if (collection != null) {
                Server target = collection.getServer();
                try {
                    target.sendCommand(command);
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(
                            null,
                            "Server " + target.getName() + " caused an exception.",
                            "Error",
                            JOptionPane.ERROR_MESSAGE
                    );
                }
            } else {
                JOptionPane.showMessageDialog(
                        null,
                        "Server " + serverName + " is invalid.",
                        "Error", JOptionPane.ERROR_MESSAGE
                );
            }
        }
    }

    /**
     * Reads all server tabs, and saves it to the variables of the corresponding servers.
     * Saves all profiles and servers to a text file.
     */
    public void save() {
        for (Collection collection : this.collections) {
            Server server = collection.getServer();
            ServerTab serverTab = collection.getServerTab();
            server.setPath(serverTab.getPath());
            server.setMaxRam(serverTab.getMaxRam());
            server.setType(ServerType.getByName(serverTab.getType()));
            try {
                server.setServerVersion(serverTab.getVersion());
            } catch (IllegalArgumentException e) {
                JOptionPane.showMessageDialog(
                        null,
                        "Invalid server version for " + server.getName(),
                        "Error",
                        JOptionPane.ERROR_MESSAGE
                );
            }
            server.toggle(serverTab.enabled());
        }
        try (PrintWriter file = new PrintWriter("files/Profiles.txt")) {
            file.println(current.name);
            file.close();
            for (Profile profile : profiles) {
                StringBuilder saveString = new StringBuilder(String.format(
                        "%s;%b;%d;%b?",
                        profile.name,
                        profile.runInBackground,
                        profile.delayStartup,
                        profile.downloadJars)
                );
                for (Collection collection : profile.collections) {
                    Server server = collection.getServer();
                    saveString.append(String.format(
                            "%s;%s;%b;%s;%s;%s;%s;%s;%s;%s!",
                            server.getName(),
                            server.getPath(),
                            server.isEnabled(),
                            server.getTypeName(),
                            server.getServerVersion(),
                            server.getMaxRam(),
                            server.getVanillaVersion(),
                            server.getSnapshotVersion(),
                            server.getSpongeVanillaVersion(),
                            server.getBungeeVersion())
                    );
                }
                saveString = new StringBuilder(saveString.substring(0, saveString.length() - 1));
                try (PrintWriter fileAppend = new PrintWriter(new FileWriter(
                        "files/Profiles.txt",
                        true
                ))) {
                    fileAppend.println(saveString);
                } catch (IOException e) {
                    JOptionPane.showMessageDialog(
                            null,
                            "Unable to save to file. Try running the software as an administrator.",
                            "Error",
                            JOptionPane.ERROR_MESSAGE
                    );
                }
            }
        } catch (IOException e) {
            JOptionPane.showMessageDialog(
                    null,
                    "Unable to save to file. Try running the software as an administrator.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
            );
        }
    }

    /**
     * Reads profiles and servers from a text file.
     */
    public static void load() {
        try (Scanner in = new Scanner(new File("files/Profiles.txt"))) {
            String profileName = in.nextLine();
            try {
                while (in.hasNextLine()) {
                    String line = in.nextLine();
                    if (line.contains("?")) {
                        String[] data = line.split("\\?");
                        String[] profileData = data[0].split(";", -1);
                        Profile profile = parseProfile(profileData);
                        Main.gui().addProfile(profileData[0]);
                        if (data[1].contains("!")) {
                            String[] servers = data[1].split("!", -1);
                            for (String server : servers) {
                                String[] serverData = server.split(";", -1);
                                parseServer(profile, serverData);
                            }
                        } else {
                            String[] serverData = data[1].split(";", -1);
                            parseServer(profile, serverData);
                        }
                    } else {
                        String[] profileData = line.split(";", -1);
                        parseProfile(profileData);
                        Main.gui().addProfile(profileData[0]);
                    }
                }
                current = getProfile(profileName);
            } catch (ArrayIndexOutOfBoundsException | NumberFormatException e) {
                e.printStackTrace();
                JOptionPane.showMessageDialog(
                        null,
                        "Invalid Profile.txt file. Profiles could not be loaded. If this error persists, please manually delete the file.",
                        "Error",
                        JOptionPane.ERROR_MESSAGE
                );
                System.exit(1);
            }
            if (profiles.size() == 0) {
                addProfile("Default");
            }
        } catch (FileNotFoundException e) {
            JOptionPane.showMessageDialog(
                    null,
                    "A profiles file was not found. Default profile was created.",
                    "Info",
                    JOptionPane.INFORMATION_MESSAGE
            );
            addProfile("Default");
        }
        Main.gui().update();
        if (current.getRunInBackground()) {
            Main.gui().hide();
            Executors.newSingleThreadExecutor().execute(Server::startServers);
        }
    }

    private static Profile parseProfile(String[] profileData) {
        return new Profile(
                profileData[0],
                Boolean.parseBoolean(profileData[1]),
                Integer.parseInt(profileData[2]),
                Boolean.parseBoolean(profileData[3])
        );
    }

    private static void parseServer(Profile profile, String[] serverData) {
        profile.collections.add(new Collection(
                serverData[0],
                serverData[1],
                Boolean.parseBoolean(serverData[2]),
                serverData[3],
                serverData[4],
                serverData[5],
                serverData[6],
                serverData[7],
                serverData[8],
                serverData[9])
        );
    }
}
